"""Data Preparation class to handle data processing for machine learning."""

__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2020-07-20'

import copy
import datetime as dt
import dill
import hashlib
import logging
import os
from functools import reduce, partial
from typing import Union, List, Iterable, Tuple, Dict, Optional

import numpy as np
import pandas as pd
import xarray as xr

from mlair.configuration import check_path_and_create
from mlair import helpers
from mlair.helpers import join, statistics, TimeTrackingWrapper, to_list
from mlair.data_handler.abstract_data_handler import AbstractDataHandler

# define a more general date type for type hinting
date = Union[dt.date, dt.datetime]
str_or_list = Union[str, List[str]]
number = Union[float, int]
num_or_list = Union[number, List[number]]
data_or_none = Union[xr.DataArray, None]


class DataHandlerSingleStation(AbstractDataHandler):
    DEFAULT_STATION_TYPE = "background"
    DEFAULT_NETWORK = "AIRBASE"
    DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                            'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                            'pblheight': 'maximum'}
    DEFAULT_WINDOW_LEAD_TIME = 3
    DEFAULT_WINDOW_HISTORY_SIZE = 13
    DEFAULT_WINDOW_HISTORY_OFFSET = 0
    DEFAULT_TIME_DIM = "datetime"
    DEFAULT_TARGET_VAR = "o3"
    DEFAULT_TARGET_DIM = "variables"
    DEFAULT_ITER_DIM = "Stations"
    DEFAULT_WINDOW_DIM = "window"
    DEFAULT_SAMPLING = "daily"
    DEFAULT_INTERPOLATION_LIMIT = 0
    DEFAULT_INTERPOLATION_METHOD = "linear"

    _hash = ["station", "statistics_per_var", "data_origin", "station_type", "network", "sampling", "target_dim",
             "target_var", "time_dim", "iter_dim", "window_dim", "window_history_size", "window_history_offset",
             "window_lead_time", "interpolation_limit", "interpolation_method"]

    def __init__(self, station, data_path, statistics_per_var, station_type=DEFAULT_STATION_TYPE,
                 network=DEFAULT_NETWORK, sampling: Union[str, Tuple[str]] = DEFAULT_SAMPLING,
                 target_dim=DEFAULT_TARGET_DIM, target_var=DEFAULT_TARGET_VAR, time_dim=DEFAULT_TIME_DIM,
                 iter_dim=DEFAULT_ITER_DIM, window_dim=DEFAULT_WINDOW_DIM,
                 window_history_size=DEFAULT_WINDOW_HISTORY_SIZE, window_history_offset=DEFAULT_WINDOW_HISTORY_OFFSET,
                 window_lead_time=DEFAULT_WINDOW_LEAD_TIME,
                 interpolation_limit: Union[int, Tuple[int]] = DEFAULT_INTERPOLATION_LIMIT,
                 interpolation_method: Union[str, Tuple[str]] = DEFAULT_INTERPOLATION_METHOD,
                 overwrite_local_data: bool = False, transformation=None, store_data_locally: bool = True,
                 min_length: int = 0, start=None, end=None, variables=None, data_origin: Dict = None,
                 lazy_preprocessing: bool = False, aggregation_dim=None, **kwargs):
        super().__init__()
        self.station = helpers.to_list(station)
        self.path = self.setup_data_path(data_path, sampling)
        self.lazy = lazy_preprocessing
        self.lazy_path = None
        if self.lazy is True:
            self.lazy_path = os.path.join(data_path, "lazy_data", self.__class__.__name__)
            check_path_and_create(self.lazy_path)
        self.statistics_per_var = statistics_per_var
        self.data_origin = data_origin
        self.do_transformation = transformation is not None
        self.input_data, self.target_data = None, None
        self._transformation = self.setup_transformation(transformation)

        self.station_type = station_type
        self.network = network
        self.sampling = sampling
        self.target_dim = target_dim
        self.target_var = target_var
        self.time_dim = time_dim
        self.iter_dim = iter_dim
        self.aggregation_dim = time_dim if aggregation_dim is None else set(to_list(aggregation_dim)+to_list(time_dim))
        self.window_dim = window_dim
        self.window_history_size = window_history_size
        self.window_history_offset = window_history_offset
        self.window_lead_time = window_lead_time

        self.interpolation_limit = interpolation_limit
        self.interpolation_method = interpolation_method

        self.overwrite_local_data = overwrite_local_data
        self.store_data_locally = store_data_locally
        self.min_length = min_length
        self.start = start
        self.end = end

        # internal
        self._data: xr.DataArray = None  # loaded raw data
        self.meta = None
        self.variables = list(statistics_per_var.keys()) if variables is None else variables
        self.history = None
        self.label = None
        self.observation = None

        # create samples
        self.setup_samples()

    def __str__(self):
        return self.station[0]

    def __len__(self):
        assert len(self.get_X()) == len(self.get_Y())
        return len(self.get_X())

    @property
    def shape(self):
        return self._data.shape, self.get_X().shape, self.get_Y().shape

    def __repr__(self):
        return f"StationPrep(station={self.station}, data_path='{self.path}', " \
               f"statistics_per_var={self.statistics_per_var}, " \
               f"station_type='{self.station_type}', network='{self.network}', " \
               f"sampling='{self.sampling}', target_dim='{self.target_dim}', target_var='{self.target_var}', " \
               f"time_dim='{self.time_dim}', window_history_size={self.window_history_size}, " \
               f"window_lead_time={self.window_lead_time}, interpolation_limit={self.interpolation_limit}, " \
               f"interpolation_method='{self.interpolation_method}', overwrite_local_data={self.overwrite_local_data})"

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim).copy()

    def get_transposed_label(self) -> xr.DataArray:
        """Return label.

        :return: label with dimensions datetime*, window*, Stations, variables.
        """
        return self.label.squeeze([self.iter_dim, self.target_dim]).transpose(self.time_dim, self.window_dim).copy()

    def get_X(self, **kwargs):
        return self.get_transposed_history()

    def get_Y(self, **kwargs):
        return self.get_transposed_label()

    def get_coordinates(self):
        coords = self.meta.loc[["station_lon", "station_lat"]].astype(float)
        return coords.rename(index={"station_lon": "lon", "station_lat": "lat"}).to_dict()[str(self)]

    def call_transform(self, inverse=False):
        opts_input = self._transformation[0]
        self.input_data, opts_input = self.transform(self.input_data, dim=self.aggregation_dim, inverse=inverse,
                                                     opts=opts_input, transformation_dim=self.target_dim)
        opts_target = self._transformation[1]
        self.target_data, opts_target = self.transform(self.target_data, dim=self.aggregation_dim, inverse=inverse,
                                                       opts=opts_target, transformation_dim=self.target_dim)
        self._transformation = (opts_input, opts_target)

    def transform(self, data_in, dim: Union[str, int] = 0, inverse: bool = False, opts=None,
                  transformation_dim=DEFAULT_TARGET_DIM):
        """
        Transform data according to given transformation settings.

        This function transforms a xarray.dataarray (along dim) or pandas.DataFrame (along axis) either with mean=0
        and std=1 (`method=standardise`) or centers the data with mean=0 and no change in data scale
        (`method=centre`). Furthermore, this sets an internal instance attribute for later inverse transformation. This
        method will raise an AssertionError if an internal transform method was already set ('inverse=False') or if the
        internal transform method, internal mean and internal standard deviation weren't set ('inverse=True').

        :param string/int dim: This param is not used for inverse transformation.
                | for xarray.DataArray as string: name of dimension which should be standardised
                | for pandas.DataFrame as int: axis of dimension which should be standardised
        :param inverse: Switch between transformation and inverse transformation.

        :return: xarray.DataArrays or pandas.DataFrames:
                #. mean: Mean of data
                #. std: Standard deviation of data
                #. data: Standardised data
        """

        def f(data, method="standardise", feature_range=None):
            if method == "standardise":
                return statistics.standardise(data, dim)
            elif method == "centre":
                return statistics.centre(data, dim)
            elif method == "min_max":
                kwargs = {"feature_range": feature_range} if feature_range is not None else {}
                return statistics.min_max(data, dim, **kwargs)
            elif method == "log":
                return statistics.log(data, dim)
            else:
                raise NotImplementedError

        def f_apply(data, method, **kwargs):
            for k, v in kwargs.items():
                if not (isinstance(v, xr.DataArray) or v is None or k == "feature_range"):
                    _, opts = statistics.min_max(data, dim)
                    helper = xr.ones_like(opts['min'])
                    kwargs[k] = helper * v
            mean = kwargs.pop('mean', None)
            std = kwargs.pop('std', None)
            min = kwargs.pop('min', None)
            max = kwargs.pop('max', None)
            feature_range = kwargs.pop('feature_range', None)

            if method == "standardise":
                return statistics.standardise_apply(data, mean, std), {"mean": mean, "std": std, "method": method}
            elif method == "centre":
                return statistics.centre_apply(data, mean), {"mean": mean, "method": method}
            elif method == "min_max":
                return statistics.min_max_apply(data, min, max), {"min": min, "max": max, "method": method,
                                                                  "feature_range": feature_range}
            elif method == "log":
                return statistics.log_apply(data, mean, std), {"mean": mean, "std": std, "method": method}
            else:
                raise NotImplementedError

        opts = opts or {}
        opts_updated = {}
        if not inverse:
            transformed_values = []
            for var in data_in.variables.values:
                data_var = data_in.sel(**{transformation_dim: [var]})
                var_opts = opts.get(var, {})
                _apply = (var_opts.get("mean", None) is not None) or (var_opts.get("min") is not None)
                values, new_var_opts = locals()["f_apply" if _apply else "f"](data_var, **var_opts)
                opts_updated[var] = copy.deepcopy(new_var_opts)
                transformed_values.append(values)
            return xr.concat(transformed_values, dim=transformation_dim), opts_updated
        else:
            return self.inverse_transform(data_in, opts, transformation_dim)

    @TimeTrackingWrapper
    def setup_samples(self):
        """
        Setup samples. This method prepares and creates samples X, and labels Y.
        """
        if self.lazy is False:
            self.make_input_target()
        else:
            self.load_lazy()
            self.store_lazy()
        if self.do_transformation is True:
            self.call_transform()
        self.make_samples()

    def store_lazy(self):
        hash = self._get_hash()
        filename = os.path.join(self.lazy_path, hash + ".pickle")
        if not os.path.exists(filename):
            dill.dump(self._create_lazy_data(), file=open(filename, "wb"))

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data]

    def load_lazy(self):
        hash = self._get_hash()
        filename = os.path.join(self.lazy_path, hash + ".pickle")
        try:
            with open(filename, "rb") as pickle_file:
                lazy_data = dill.load(pickle_file)
            self._extract_lazy(lazy_data)
            logging.debug(f"{self.station[0]}: used lazy data")
        except FileNotFoundError:
            logging.debug(f"{self.station[0]}: could not use lazy data")
            self.make_input_target()

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data, self.input_data, self.target_data = list(map(f_prep, [_data, _input_data, _target_data]))

    def make_input_target(self):
        data, self.meta = self.load_data(self.path, self.station, self.statistics_per_var, self.sampling,
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         self.start, self.end)
        self._data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
                                      limit=self.interpolation_limit, sampling=self.sampling)
        self.set_inputs_and_targets()

    def set_inputs_and_targets(self):
        inputs = self._data.sel({self.target_dim: helpers.to_list(self.variables)})
        targets = self._data.sel(
            {self.target_dim: helpers.to_list(self.target_var)})  # ToDo: is it right to expand this dim??
        self.input_data = inputs
        self.target_data = targets

    def make_samples(self):
        self.make_history_window(self.target_dim, self.window_history_size, self.time_dim)
        self.make_labels(self.target_dim, self.target_var, self.time_dim, self.window_lead_time)
        self.make_observation(self.target_dim, self.target_var, self.time_dim)
        self.remove_nan(self.time_dim)

    def load_data(self, path, station, statistics_per_var, sampling, station_type=None, network=None,
                  store_data_locally=False, data_origin: Dict = None, start=None, end=None):
        """
        Load data and meta data either from local disk (preferred) or download new data by using a custom download method.

        Data is either downloaded, if no local data is available or parameter overwrite_local_data is true. In both
        cases, downloaded data is only stored locally if store_data_locally is not disabled. If this parameter is not
        set, it is assumed, that data should be saved locally.
        """
        check_path_and_create(path)
        file_name = self._set_file_name(path, station, statistics_per_var)
        meta_file = self._set_meta_file_name(path, station, statistics_per_var)
        if self.overwrite_local_data is True:
            logging.debug(f"overwrite_local_data is true, therefore reload {file_name}")
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists(meta_file):
                os.remove(meta_file)
            data, meta = self.download_data(file_name, meta_file, station, statistics_per_var, sampling,
                                            station_type=station_type, network=network,
                                            store_data_locally=store_data_locally, data_origin=data_origin)
            logging.debug(f"loaded new data")
        else:
            try:
                logging.debug(f"try to load local data from: {file_name}")
                data = xr.open_dataarray(file_name)
                meta = pd.read_csv(meta_file, index_col=0)
                self.check_station_meta(meta, station, station_type, network)
                logging.debug("loading finished")
            except FileNotFoundError as e:
                logging.debug(e)
                logging.debug(f"load new data")
                data, meta = self.download_data(file_name, meta_file, station, statistics_per_var, sampling,
                                                station_type=station_type, network=network,
                                                store_data_locally=store_data_locally, data_origin=data_origin)
                logging.debug("loading finished")
        # create slices and check for negative concentration.
        data = self._slice_prep(data, start=start, end=end)
        data = self.check_for_negative_concentrations(data)
        return data, meta

    @staticmethod
    def download_data_from_join(file_name: str, meta_file: str, station, statistics_per_var, sampling,
                                station_type=None, network=None, store_data_locally=True, data_origin: Dict = None,
                                time_dim=DEFAULT_TIME_DIM, target_dim=DEFAULT_TARGET_DIM, iter_dim=DEFAULT_ITER_DIM) \
            -> [xr.DataArray, pd.DataFrame]:
        """
        Download data from TOAR database using the JOIN interface.

        Data is transformed to a xarray dataset. If class attribute store_data_locally is true, data is additionally
        stored locally using given names for file and meta file.

        :param file_name: name of file to save data to (containing full path)
        :param meta_file: name of the meta data file (also containing full path)

        :return: downloaded data and its meta data
        """
        df_all = {}
        df, meta = join.download_join(station_name=station, stat_var=statistics_per_var, station_type=station_type,
                                      network_name=network, sampling=sampling, data_origin=data_origin)
        df_all[station[0]] = df
        # convert df_all to xarray
        xarr = {k: xr.DataArray(v, dims=[time_dim, target_dim]) for k, v in df_all.items()}
        xarr = xr.Dataset(xarr).to_array(dim=iter_dim)
        if store_data_locally is True:
            # save locally as nc/csv file
            xarr.to_netcdf(path=file_name)
            meta.to_csv(meta_file)
        return xarr, meta

    def download_data(self, *args, **kwargs):
        data, meta = self.download_data_from_join(*args, **kwargs, time_dim=self.time_dim, target_dim=self.target_dim,
                                                  iter_dim=self.iter_dim)
        return data, meta

    @staticmethod
    def check_station_meta(meta, station, station_type, network):
        """
        Search for the entries in meta data and compare the value with the requested values.

        Will raise a FileNotFoundError if the values mismatch.
        """
        if station_type is not None:
            check_dict = {"station_type": station_type, "network_name": network}
            for (k, v) in check_dict.items():
                if v is None:
                    continue
                if meta.at[k, station[0]] != v:
                    logging.debug(f"meta data does not agree with given request for {k}: {v} (requested) != "
                                  f"{meta.at[k, station[0]]} (local). Raise FileNotFoundError to trigger new "
                                  f"grapping from web.")
                    raise FileNotFoundError

    def check_for_negative_concentrations(self, data: xr.DataArray, minimum: int = 0) -> xr.DataArray:
        """
        Set all negative concentrations to zero.

        Names of all concentrations are extracted from https://join.fz-juelich.de/services/rest/surfacedata/
        #2.1 Parameters. Currently, this check is applied on "benzene", "ch4", "co", "ethane", "no", "no2", "nox",
        "o3", "ox", "pm1", "pm10", "pm2p5", "propane", "so2", and "toluene".

        :param data: data array containing variables to check
        :param minimum: minimum value, by default this should be 0

        :return: corrected data
        """
        chem_vars = ["benzene", "ch4", "co", "ethane", "no", "no2", "nox", "o3", "ox", "pm1", "pm10", "pm2p5",
                     "propane", "so2", "toluene"]
        # used_chem_vars = list(set(chem_vars) & set(self.statistics_per_var.keys()))
        used_chem_vars = list(set(chem_vars) & set(data.variables.values))
        if len(used_chem_vars) > 0:
            data.loc[..., used_chem_vars] = data.loc[..., used_chem_vars].clip(min=minimum)
        return data

    def setup_data_path(self, data_path: str, sampling: str):
        return os.path.join(os.path.abspath(data_path), sampling)

    def shift(self, data: xr.DataArray, dim: str, window: int, offset: int = 0) -> xr.DataArray:
        """
        Shift data multiple times to represent history (if window <= 0) or lead time (if window > 0).

        :param data: data set to shift
        :param dim: dimension along shift is applied
        :param window: number of steps to shift (corresponds to the window length)
        :param offset: use offset to move the window by as many time steps as given in offset. This can be used, if the
            index time of a history element is not the last timestamp. E.g. you could use offset=23 when dealing with
            hourly data in combination with daily data (values from 00 to 23 are aggregated on 00 the same day).

        :return: shifted data
        """
        start = 1
        end = 1
        if window <= 0:
            start = window
        else:
            end = window + 1
        res = []
        _range = list(map(lambda x: x + offset, range(start, end)))
        for w in _range:
            res.append(data.shift({dim: -w}))
        window_array = self.create_index_array(self.window_dim, _range, squeeze_dim=self.target_dim)
        res = xr.concat(res, dim=window_array)
        return res

    @staticmethod
    def create_index_array(index_name: str, index_value: Iterable[int], squeeze_dim: str) -> xr.DataArray:
        """
        Create an 1D xr.DataArray with given index name and value.

        :param index_name: name of dimension
        :param index_value: values of this dimension

        :return: this array
        """
        ind = pd.DataFrame({'val': index_value}, index=index_value)
        # res = xr.Dataset.from_dataframe(ind).to_array().rename({'index': index_name}).squeeze(dim=squeez/e_dim, drop=True)
        res = xr.Dataset.from_dataframe(ind).to_array(squeeze_dim).rename({'index': index_name}).squeeze(
            dim=squeeze_dim,
            drop=True
        )
        res.name = index_name
        return res

    @staticmethod
    def _set_file_name(path, station, statistics_per_var):
        all_vars = sorted(statistics_per_var.keys())
        return os.path.join(path, f"{''.join(station)}_{'_'.join(all_vars)}.nc")

    @staticmethod
    def _set_meta_file_name(path, station, statistics_per_var):
        all_vars = sorted(statistics_per_var.keys())
        return os.path.join(path, f"{''.join(station)}_{'_'.join(all_vars)}_meta.csv")

    def interpolate(self, data, dim: str, method: str = 'linear', limit: int = None,
                    use_coordinate: Union[bool, str] = True, sampling="daily", **kwargs):
        """
        Interpolate values according to different methods.

        (Copy paste from dataarray.interpolate_na)

        :param dim:
                Specifies the dimension along which to interpolate.
        :param method:
                {'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                          'polynomial', 'barycentric', 'krog', 'pchip',
                          'spline', 'akima'}, optional
                    String indicating which method to use for interpolation:

                    - 'linear': linear interpolation (Default). Additional keyword
                      arguments are passed to ``numpy.interp``
                    - 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                      'polynomial': are passed to ``scipy.interpolate.interp1d``. If
                      method=='polynomial', the ``order`` keyword argument must also be
                      provided.
                    - 'barycentric', 'krog', 'pchip', 'spline', and `akima`: use their
                      respective``scipy.interpolate`` classes.
        :param limit:
                    default None
                    Maximum number of consecutive NaNs to fill. Must be greater than 0
                    or None for no limit.
        :param use_coordinate:
                default True
                    Specifies which index to use as the x values in the interpolation
                    formulated as `y = f(x)`. If False, values are treated as if
                    eqaully-spaced along `dim`. If True, the IndexVariable `dim` is
                    used. If use_coordinate is a string, it specifies the name of a
                    coordinate variariable to use as the index.
        :param kwargs:

        :return: xarray.DataArray
        """
        data = self.create_full_time_dim(data, dim, sampling)
        return data.interpolate_na(dim=dim, method=method, limit=limit, use_coordinate=use_coordinate, **kwargs)

    @staticmethod
    def create_full_time_dim(data, dim, sampling):
        """Ensure time dimension to be equidistant. Sometimes dates if missing values have been dropped."""
        start = data.coords[dim].values[0]
        end = data.coords[dim].values[-1]
        freq = {"daily": "1D", "hourly": "1H"}.get(sampling)
        datetime_index = pd.DataFrame(index=pd.date_range(start, end, freq=freq))
        t = data.sel({dim: start}, drop=True)
        res = xr.DataArray(coords=[datetime_index.index, *[t.coords[c] for c in t.coords]], dims=[dim, *t.coords])
        res = res.transpose(*data.dims)
        res.loc[data.coords] = data
        return res

    def make_history_window(self, dim_name_of_inputs: str, window: int, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing history data.

        Shift the data window+1 times and return a xarray which has a new dimension 'window' containing the shifted
        data. This is used to represent history in the data. Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: number of time steps to look back in history
                Note: window will be treated as negative value. This should be in agreement with looking back on
                a time line. Nonetheless positive values are allowed but they are converted to its negative
                expression
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        history = self._make_and_return_history_window(dim_name_of_shift, window)
        self.history = history
        # self.history = self.shift(data, dim_name_of_shift, window, offset=self.window_history_offset)

    def _make_and_return_history_window(self, dim_name_of_shift, window, data=None):
        """
        Create a xr.DataArray containing history data.

        Shift the data window+1 times and return a xarray which has a new dimension 'window' containing the shifted
        data. This is used to represent history in the data. Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: number of time steps to look back in history
                Note: window will be treated as negative value. This should be in agreement with looking back on
                a time line. Nonetheless positive values are allowed but they are converted to its negative
                expression
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        if data is None:
            data = self.input_data
        window = -abs(window)
        history = self.shift(data, dim_name_of_shift, window, offset=self.window_history_offset)
        return history

    def make_labels(self, dim_name_of_target: str, target_var: str_or_list, dim_name_of_shift: str,
                    window: int) -> None:
        """
        Create a xr.DataArray containing labels.

        Labels are defined as the consecutive target values (t+1, ...t+n) following the current time step t. Set label
        attribute.

        :param dim_name_of_target: Name of dimension which contains the target variable
        :param target_var: Name of target variable in 'dimension'
        :param dim_name_of_shift: Name of dimension on which xarray.DataArray.shift will be applied
        :param window: lead time of label
        """
        window = abs(window)
        data = self.target_data
        self.label = self.shift(data, dim_name_of_shift, window)

    def make_observation(self, dim_name_of_target: str, target_var: str_or_list, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing observations.

        Observations are defined as value of the current time step t. Set observation attribute.

        :param dim_name_of_target: Name of dimension which contains the observation variable
        :param target_var: Name of observation variable(s) in 'dimension'
        :param dim_name_of_shift: Name of dimension on which xarray.DataArray.shift will be applied
        """
        data = self.target_data
        self.observation = self.shift(data, dim_name_of_shift, 0)

    def remove_nan(self, dim: str) -> None:
        """
        Remove all NAs slices along dim which contain nans in history, label and observation.

        This is done to present only a full matrix to keras.fit. Update history, label, and observation attribute.

        :param dim: dimension along the remove is performed.
        """
        intersect = []
        if (self.history is not None) and (self.label is not None):
            non_nan_history = self.history.dropna(dim=dim)
            non_nan_label = self.label.dropna(dim=dim)
            non_nan_observation = self.observation.dropna(dim=dim)
            intersect = reduce(np.intersect1d, (non_nan_history.coords[dim].values, non_nan_label.coords[dim].values,
                                                non_nan_observation.coords[dim].values))

        if len(intersect) < max(self.min_length, 1):
            self.history = None
            self.label = None
            self.observation = None
        else:
            self.history = self.history.sel({dim: intersect})
            self.label = self.label.sel({dim: intersect})
            self.observation = self.observation.sel({dim: intersect})

    def _slice_prep(self, data: xr.DataArray, start=None, end=None) -> xr.DataArray:
        """
        Set start and end date for slicing and execute self._slice().

        :param data: data to slice
        :param coord: name of axis to slice

        :return: sliced data
        """
        start = start if start is not None else data.coords[self.time_dim][0].values
        end = end if end is not None else data.coords[self.time_dim][-1].values
        return self._slice(data, start, end, self.time_dim)

    @staticmethod
    def _slice(data: xr.DataArray, start: Union[date, str], end: Union[date, str], coord: str) -> xr.DataArray:
        """
        Slice through a given data_item (for example select only values of 2011).

        :param data: data to slice
        :param start: start date of slice
        :param end: end date of slice
        :param coord: name of axis to slice

        :return: sliced data
        """
        return data.loc[{coord: slice(str(start), str(end))}]

    def setup_transformation(self, transformation: Union[None, dict, Tuple]) -> Tuple[Optional[dict], Optional[dict]]:
        """
        Set up transformation by extracting all relevant information.

        * Either return new empty DataClass instances if given transformation arg is None,
        * or return given object twice if transformation is a DataClass instance,
        * or return the inputs and targets attributes if transformation is a TransformationClass instance (default
          design behaviour)
        """
        if transformation is None:
            return None, None
        elif isinstance(transformation, dict):
            return copy.deepcopy(transformation), copy.deepcopy(transformation)
        elif isinstance(transformation, tuple) and len(transformation) == 2:
            return copy.deepcopy(transformation)
        else:
            raise NotImplementedError("Cannot handle this.")

    @staticmethod
    def check_inverse_transform_params(method: str, mean=None, std=None, min=None, max=None) -> None:
        """
        Support inverse_transformation method.

        Validate if all required statistics are available for given method. E.g. centering requires mean only, whereas
        normalisation requires mean and standard deviation. Will raise an AttributeError on missing requirements.

        :param mean: data with all mean values
        :param std: data with all standard deviation values
        :param method: name of transformation method
        """
        msg = ""
        if method in ['standardise', 'centre'] and mean is None:
            msg += "mean, "
        if method == 'standardise' and std is None:
            msg += "std, "
        if method == "min_max" and min is None:
            msg += "min, "
        if method == "min_max" and max is None:
            msg += "max, "
        if len(msg) > 0:
            raise AttributeError(f"Inverse transform {method} can not be executed because following is None: {msg}")

    def inverse_transform(self, data_in, opts, transformation_dim) -> xr.DataArray:
        """
        Perform inverse transformation.

        Will raise an AssertionError, if no transformation was performed before. Checks first, if all required
        statistics are available for inverse transformation. Class attributes data, mean and std are overwritten by
        new data afterwards. Thereby, mean, std, and the private transform method are set to None to indicate, that the
        current data is not transformed.
        """

        def f_inverse(data, method, mean=None, std=None, min=None, max=None, feature_range=None):
            if method == "standardise":
                return statistics.standardise_inverse(data, mean, std)
            elif method == "centre":
                return statistics.centre_inverse(data, mean)
            elif method == "min_max":
                return statistics.min_max_inverse(data, min, max, feature_range)
            elif method == "log":
                return statistics.log_inverse(data, mean, std)
            else:
                raise NotImplementedError

        transformed_values = []
        squeeze = False
        if transformation_dim in data_in.coords:
            if transformation_dim not in data_in.dims:
                data_in = data_in.expand_dims(transformation_dim)
                squeeze = True
        else:
            raise IndexError(f"Could not find given dimension: {transformation_dim}. Available is: {data_in.coords}")
        for var in data_in.variables.values:
            data_var = data_in.sel(**{transformation_dim: [var]})
            var_opts = opts.get(var, {})
            _method = var_opts.get("method", None)
            if _method is None:
                raise AssertionError(f"Inverse transformation method is not set for {var}.")
            self.check_inverse_transform_params(**var_opts)
            values = f_inverse(data_var, **var_opts)
            transformed_values.append(values)
        res = xr.concat(transformed_values, dim=transformation_dim)
        return res.squeeze(transformation_dim) if squeeze else res

    def apply_transformation(self, data, base=None, dim=0, inverse=False):
        """
        Apply transformation on external data. Specify if transformation should be based on parameters related to input
        or target data using `base`. This method can also apply inverse transformation.

        :param data:
        :param base:
        :param dim:
        :param inverse:
        :return:
        """
        pos = self.get_transformation_base(base)
        return self.transform(data, dim=dim, opts=self._transformation[pos], inverse=inverse,
                              transformation_dim=self.target_dim)

    @staticmethod
    def get_transformation_base(base):
        if base in ["target", 1]:
            pos = 1
        elif base in ["input", 0]:
            pos = 0
        else:
            raise ValueError("apply transformation requires a reference for transformation options. Please specify if"
                             "you want to use input or target transformation using the parameter 'base'. Given was: " +
                             base)
        return pos

    def _hash_list(self):
        return sorted(list(set(self._hash)))

    def _get_hash(self):
        hash = "".join([str(self.__getattribute__(e)) for e in self._hash_list()]).encode()
        return hashlib.md5(hash).hexdigest()


if __name__ == "__main__":
    # dp = AbstractDataPrep('data/', 'dummy', 'DEBW107', ['o3', 'temp'], statistics_per_var={'o3': 'dma8eu', 'temp': 'maximum'})
    # print(dp)
    statistics_per_var = {'o3': 'dma8eu', 'temp-rea-miub': 'maximum'}
    sp = DataHandlerSingleStation(data_path='/home/felix/PycharmProjects/mlt_new/data/', station='DEBY122',
                                  statistics_per_var=statistics_per_var, station_type='background',
                                  network='UBA', sampling='daily', target_dim='variables', target_var='o3',
                                  time_dim='datetime', window_history_size=7, window_lead_time=3,
                                  interpolation_limit=0
                                  )  # transformation={'method': 'standardise'})
    sp2 = DataHandlerSingleStation(data_path='/home/felix/PycharmProjects/mlt_new/data/', station='DEBY122',
                                   statistics_per_var=statistics_per_var, station_type='background',
                                   network='UBA', sampling='daily', target_dim='variables', target_var='o3',
                                   time_dim='datetime', window_history_size=7, window_lead_time=3,
                                   transformation={'method': 'standardise'})
    sp2.transform(inverse=True)
    sp.get_X()
    sp.get_Y()
    print(len(sp))
    print(sp.shape)
    print(sp)
